#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <time.h>
#include <direct.h>
#include "sticky.h"

using namespace std;

string rootdir;

void
loadfile(const char *filename)
{
    string s;
    char buf[8192];

    FILE *fp = fopen(filename, "r");
    fgets(buf, sizeof(buf), fp);
    int x = atoi(strtok(buf, ", "));
    int y = atoi(strtok(NULL, ", "));
    int w = atoi(strtok(NULL, ", "));
    int h = atoi(strtok(NULL, ", "));

    if (!x) x = 100;
    if (!y) y = 100;
    if (!w) w = 200;
    if (!h) h = 100;

    while (fgets(buf, sizeof(buf), fp)) {
	int n = strlen(buf);
	if (buf[n-1] == '\n') {
	    if (n) buf[n-1] = 0;
	    s += buf;
	    s += "\r\n";
	} else
	    s += buf;
    }
    fclose(fp);

//    printf("<%s> %d, %d %dx%d %s\n", filename, x, y, w, h, s.c_str());
//    fflush(stdout);

    StickyNote *sn = new_note(x, y, w, h, s.c_str());
    sn->filename = filename;
}

void
load(void)
{
    string s = rootdir;

    WIN32_FIND_DATA dt;
    HANDLE dir;
    string dirglob(s.c_str());
    dirglob += "\\*";

    dir = FindFirstFile(dirglob.c_str(), &dt);
    while (dir != INVALID_HANDLE_VALUE) {
	string fullpath;
	fullpath = rootdir;
	fullpath += "\\";
	fullpath += dt.cFileName;
	struct stat st;

	/* skip if we can't stat() it */
	if (stat(fullpath.c_str(), &st) >= 0) {
	    if (((st.st_mode) & S_IFMT) == S_IFREG) {
		if (!strncmp(dt.cFileName, "Note_", 5)) {
		    loadfile(fullpath.c_str());
		}
	    }
	}
	if (!FindNextFile(dir, &dt))
	    break;
    }
    FindClose(dir);
}

void
initcfg(void)
{
    bool err = 0;
    HKEY rk;

    if (RegOpenKeyEx(HKEY_CURRENT_USER, "Volatile Environment", 0,
		       KEY_ALL_ACCESS, &rk) != ERROR_SUCCESS)
	err = 1;

    char buf[8192];
    buf[0] = 0;
    DWORD sz = sizeof(buf);
    DWORD ty = REG_SZ;

    if (!err && RegQueryValueEx(rk, "APPDATA", NULL, &ty, (LPBYTE)buf, &sz) != ERROR_SUCCESS)
	err = 1;

    RegCloseKey(rk);

    if (err) {
	MessageBox(NULL, NULL, "FATAL: Could not location application data dir",
		   MB_ICONWARNING | MB_OK);
	PostQuitMessage(0);
	return;
    }

    rootdir.assign(buf, sz-1);
    rootdir.append("\\Ishiboo");
    mkdir(rootdir.c_str());
    rootdir.append("\\Sticky Notes");
    mkdir(rootdir.c_str());
}

void
save(StickyNote *sn)
{
    string s = rootdir;

    WINDOWPLACEMENT wp;
    wp.length = sizeof(wp);
    GetWindowPlacement(sn->hwnd, &wp);

    int x = wp.rcNormalPosition.left;
    int y = wp.rcNormalPosition.top;
    int w = wp.rcNormalPosition.right - wp.rcNormalPosition.left;
    int h = wp.rcNormalPosition.bottom - wp.rcNormalPosition.top;

    wp.length = sizeof(wp);
    GetWindowPlacement(GetDesktopWindow(), &wp);

    int ww = wp.rcNormalPosition.right - wp.rcNormalPosition.left;
    int hh = wp.rcNormalPosition.bottom - wp.rcNormalPosition.top;
    while (x < 0) x += ww;
    while (y < 0) y += hh;

    char buf[100];

    if (sn->filename.empty()) {
	sn->filename = s;
	sn->filename += "\\Note_";

	sprintf(buf, "%x-%x", time(NULL), g_seq++);
	sn->filename += buf;
    }

    FILE *fp = fopen(sn->filename.c_str(), "wb");
    char *b;
    int i = SendDlgItemMessage(sn->hwnd, 1, WM_GETTEXTLENGTH, 0, 0);
    b = new char[i+1];
    SendDlgItemMessage(sn->hwnd, 1, WM_GETTEXT, i+1, (LPARAM)b);
    fprintf(fp, "%d,%d,%d,%d\r\n%s", x, y, w, h, b);
    delete[] b;
    fclose(fp);

//    printf("saving <%s>\n", sn->filename.c_str());
//    fflush(stdout);
}
