// C4800: 'int' : forcing value to bool 'true' or 'false' (performance warning)
#pragma warning(disable : 4800)
// Win32 templates generate hundreds of "overly-long identifier" warnings
#pragma warning(disable : 4786)
#pragma warning(disable : 4503)

int g_seq = 0;

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#include <set>
#include "resource.h"

#include "sticky.h"
#include <richedit.h>

using namespace std;

// -------------------------------------

set<StickyNote *> g_stickies;

HBRUSH g_bg_brush;

void
deletenote(HWND hwnd)
{
    StickyNote *sn = (StickyNote*)GetProp(hwnd, "ctrldata");
    g_stickies.erase(sn);
    DestroyWindow(hwnd);
    unlink(sn->filename.c_str());
    delete sn;
}

LRESULT CALLBACK
NotifyWndProc2(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    StickyNote *sn = (StickyNote*)GetProp(hwnd, "ctrldata");
    static int x, y;
    static bool moving = 0;
    HWND h;
    switch (msg) {
	case WM_CREATE:
	    {
		sn = (StickyNote*)(((CREATESTRUCT*)lParam)->lpCreateParams);
		SetProp(hwnd, "ctrldata", (HANDLE)sn);

		BringWindowToTop(hwnd);

		// this is the rich edit control
		h = CreateWindow(RICHEDIT_CLASS, NULL,
				 WS_CHILD | WS_VISIBLE | WS_VSCROLL |
				 ES_LEFT | ES_MULTILINE | ES_AUTOVSCROLL |
				 ES_WANTRETURN,
				 0, 0, 0, 0, hwnd,
				 (HMENU)1, GetModuleHandle(NULL), NULL);

		// fix the font
		SendMessage(h, WM_SETFONT, (WPARAM)GetStockObject(DEFAULT_GUI_FONT), 1);

		// get change and click events
		SendMessage(h, EM_SETEVENTMASK, 0, ENM_CHANGE|ENM_LINK);

		// set bg to yellow
		SendMessage(h, EM_SETBKGNDCOLOR, 0, (LPARAM)RGB(0xff, 0xff, 0));

		// auto detect the urls typed into the control
		SendMessage(h, EM_AUTOURLDETECT, 1, 0);

		// set the text we had saved. do this after the url detection
		// or else urls won't be detected
		SendMessage(h, WM_SETTEXT,
			    (WPARAM)0, (LPARAM)sn->data.c_str());
	    }
	    break;

	case WM_RBUTTONDOWN:
	    {
		POINT pnt;
		GetCursorPos(&pnt);
		HMENU menu = CreatePopupMenu();
		AppendMenu(menu, MF_STRING, 1, "Remove This Note");
		AppendMenu(menu, MF_SEPARATOR, 0, NULL);
		AppendMenu(menu, MF_STRING, 2, "Create New Note");
		int ret = TrackPopupMenuEx(menu,
					   TPM_LEFTALIGN | TPM_TOPALIGN | TPM_RETURNCMD,
					   pnt.x, pnt.y, hwnd, NULL);
		DestroyMenu(menu);
		if (ret == 1) {
		    deletenote(hwnd);
		    return 0;
		} else if (ret == 2) {
		    new_note(0, 0, 0, 0, NULL);
		}
	    }
	    break;

	case WM_NOTIFY:
	    if (LOWORD(wParam) == 1) {
		ENLINK *enlink = (ENLINK*)lParam;
		if (enlink->msg == WM_LBUTTONDOWN) {
		    // select the text
		    SendDlgItemMessage(hwnd, 1, EM_EXSETSEL, 0, (LPARAM)&enlink->chrg);

		    // allocate a buffer for the text
		    char *buf = new char[enlink->chrg.cpMax - enlink->chrg.cpMin + 1];

		    // get the text
		    SendDlgItemMessage(hwnd, 1, EM_GETSELTEXT, 0, (LPARAM)buf);

		    // execute the link -- use new browser if shift being held
		    if (GetAsyncKeyState(VK_SHIFT) >> 15)
			ShellExecute(hwnd, "open", "iexplore", buf, "", SW_SHOWNORMAL);
		    else
			ShellExecute(hwnd, "open", buf, "", "", SW_SHOWNORMAL);

		    delete buf;
		    return 0;
		}
	    }
	    break;

	case WM_COMMAND:
	    switch (HIWORD(wParam)) {
		case EN_CHANGE:
		    KillTimer(hwnd, 1025);
		    SetTimer(hwnd, 1025, 250, NULL);
		    break;
	    }
	    break;

	case WM_TIMER:
	    KillTimer(hwnd, 1025);
	    save(sn);
	    break;

	case WM_LBUTTONUP:
	    ReleaseCapture();
	    moving = 0;
	    break;

	case WM_LBUTTONDOWN:
	    moving = 1;
	    SetCapture(hwnd);
	    x = GET_X_LPARAM(lParam);
	    y = GET_Y_LPARAM(lParam);
	    break;

	case WM_MOUSEMOVE:
	    {
		if (moving) {
		    WINDOWPLACEMENT wp;
		    wp.length = sizeof(wp);
		    GetWindowPlacement(hwnd, &wp);

		    int xx = GET_X_LPARAM(lParam) - x;
		    int yy = GET_Y_LPARAM(lParam) - y;
		    if (xx || yy) {
			SetWindowPos(hwnd, 0,
				     xx + wp.rcNormalPosition.left,
				     yy + wp.rcNormalPosition.top,
				     0, 0,
				     SWP_NOOWNERZORDER | SWP_NOSIZE |
				     SWP_NOZORDER);
			x = GET_X_LPARAM(lParam) - xx;
			y = GET_Y_LPARAM(lParam) - yy;
		    }
		}
	    }
	    break;

	case WM_MOVE:
	    KillTimer(hwnd, 1025);
	    SetTimer(hwnd, 1025, 250, NULL);
	    break;

	case WM_SIZE:
	    {
		RECT sz;
		GetClientRect(hwnd, &sz);
		h = GetDlgItem(hwnd, 1);
		SetWindowPos(h, HWND_TOP, 0, 14, sz.right, sz.bottom - 14, 0);
		KillTimer(hwnd, 1025);
		SetTimer(hwnd, 1025, 250, NULL);
	    }
	    break;
	case WM_CLOSE:
	    return true;
    }

    return DefWindowProc(hwnd, msg, wParam, lParam);
}

StickyNote *
new_note(int x, int y, int w, int h, const char *data)
{
    static WNDCLASSEX wc;
    ZeroMemory(&wc, sizeof(wc));

    if (!GetClassInfoEx(GetModuleHandle(NULL), "sticky note", &wc)) {
	wc.style         = 0;
	wc.cbClsExtra    = 0;
	wc.cbWndExtra    = 0;
	wc.hInstance     = GetModuleHandle(NULL);
	wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
	wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
	wc.hbrBackground = (HBRUSH)g_bg_brush;
	wc.lpszMenuName  = NULL;
	wc.lpszClassName = "sticky note";
	wc.cbSize        = sizeof(wc);
	wc.lpfnWndProc   = NotifyWndProc2;
	ATOM a = RegisterClassEx(&wc);
	assert(a);
    }

    if (data == NULL) {
	WINDOWPLACEMENT wp;
	wp.length = sizeof(wp);
	GetWindowPlacement(GetDesktopWindow(), &wp);

	w = 130;
	h = 80;
	x = ((wp.rcNormalPosition.right -
	      wp.rcNormalPosition.left) / 2) - (w / 2);

	y = ((wp.rcNormalPosition.bottom -
	      wp.rcNormalPosition.top) / 2) - (h / 2);
    }

    StickyNote *sn = new StickyNote;
    g_stickies.insert(sn);
    if (data) sn->data = data;
    sn->hwnd = CreateWindowEx(WS_EX_TOOLWINDOW | WS_EX_STATICEDGE | WS_EX_TOPMOST,
			      "sticky note", " ",
			      WS_POPUP | WS_THICKFRAME | WS_VISIBLE,
			      x, y, w, h,
			      NULL, NULL, GetModuleHandle(NULL), sn);
    return sn;
}

LRESULT CALLBACK
NotifyWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    switch (msg) {
	case WM_CREATE:
	    {
		NOTIFYICONDATA tnid;
		tnid.cbSize = sizeof(NOTIFYICONDATA);
		tnid.hWnd = hwnd;
		tnid.uID = 1025;
		tnid.uFlags = NIF_ICON | NIF_TIP | NIF_MESSAGE;
		tnid.uCallbackMessage = WM_USER + 1;
		tnid.hIcon = LoadIcon(GetModuleHandle(NULL),
				      MAKEINTRESOURCE(IDI_TRAYICON));
		const char *tip = "Sticky Notes";
		lstrcpyn(tnid.szTip, tip, strlen(tip)+1);
		Shell_NotifyIcon(NIM_ADD, &tnid);
	    }
	    initcfg();
	    load();
	    break;

	case WM_USER + 1:
	    if (msg == WM_USER + 1) {
		if (LOWORD(lParam) == WM_LBUTTONDOWN) {
		    new_note(0, 0, 0, 0, NULL);
		} else if (LOWORD(lParam) == WM_RBUTTONDOWN) {
		    HMENU menu = CreatePopupMenu();
		    POINT pnt;
		    GetCursorPos(&pnt);

		    AppendMenu(menu, MF_STRING, 1, "New Sticky Note");
//		    AppendMenu(menu, MF_STRING, 2, "Settings...");
		    AppendMenu(menu, MF_STRING, 3, "Force Save All");
		    AppendMenu(menu, MF_SEPARATOR, 0, NULL);
		    AppendMenu(menu, MF_STRING, 4, "Delete All Notes");
		    AppendMenu(menu, MF_SEPARATOR, 0, NULL);
		    AppendMenu(menu, MF_STRING, 9, "Exit");

		    SetForegroundWindow(hwnd);
		    int ret = TrackPopupMenuEx(menu,
					       TPM_LEFTALIGN | TPM_TOPALIGN |
					       TPM_RETURNCMD,
					       pnt.x, pnt.y, hwnd, NULL);
		    PostMessage(hwnd, WM_NULL, 0, 0);
		    DestroyMenu(menu);

		    switch (ret) {
			case 1:
			    new_note(0, 0, 0, 0, NULL);
			    break;
			case 2:
			    // settings
			    break;
			case 3:
			    {
				set<StickyNote *>::iterator it = g_stickies.begin();
				while (it != g_stickies.end()) {
				    save(*it);
				    it++;
				}
			    }
			    break;
			case 4:
			    while (!g_stickies.empty()) {
				StickyNote *sn = *g_stickies.begin();
				deletenote(sn->hwnd);
			    }
			    break;
			case 9:
			    {
				NOTIFYICONDATA tnid;
				tnid.cbSize = sizeof(NOTIFYICONDATA);
				tnid.hWnd = hwnd;
				tnid.uID = 1025;
				Shell_NotifyIcon(NIM_DELETE, &tnid);
				PostQuitMessage(0);
			    }
			    break;
			default:
			    break;
		    }
		}
	    }
	    break;
    }
    return DefWindowProc(hwnd, msg, wParam, lParam);
}

int APIENTRY
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
	LPSTR lpCmdLine, int nCmdShow)
{
    g_bg_brush = CreateSolidBrush(RGB(0xff,0xff,0xa0));

    LoadLibrary("RICHED20.DLL");

    HWND hwnd = 0;
    {
	WNDCLASSEX wc;
	ZeroMemory(&wc, sizeof(wc));

	if (!GetClassInfoEx(hInstance, "sticky", &wc)) {
	    wc.style         = 0;
	    wc.cbClsExtra    = 0;
	    wc.cbWndExtra    = 0;
	    wc.hInstance     = hInstance;
	    wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
	    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
	    wc.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	    wc.lpszMenuName  = NULL;
	    wc.lpszClassName = "sticky";
	    wc.cbSize        = sizeof(wc);
	    wc.lpfnWndProc   = NotifyWndProc;
	    ATOM a = RegisterClassEx(&wc);
	    assert(a);
	}
	hwnd = CreateWindow("sticky", "sticky",
			    WS_OVERLAPPED, 0, 0, 100, 100,
			    NULL, NULL, hInstance, NULL);
    }

    MSG msg;
    while (GetMessage(&msg, NULL, 0, 0)) {
	TranslateMessage(&msg);
	DispatchMessage(&msg);
    }

    return 0;
}
